:: ============================================================
:: Mappings Library - A Windows batch script helper for managing key:value mappings
::
:: The file format it understands is a subset of YAML (yaml.org)
::
::   e.g. key: Some value
::
:: Usage: libmap.cmd ACTION [OPTIONS...] [PARAMETERS...]
::
:: Global options:
::   -v = verbose -- display all output and messages (info, error, etc.)
::
:: Supported actions:
::   - list (group, [separator="=>"]) -- outputs the specified group of key, value pairs
::   - get (group, key, value) -- retrieves the value associated with a key
::   - set (group, key, value) -- maps the given key to a value
::   - delete (group, key) -- removes a pair from a group based on the given key
::   - save (group, data_file_path, [create_file="True"]) -- writes the specified group of mappings to the given file
::   - load (group, data_file_path) -- loads mappings for group from given data file
::   - clear (group) -- clears all the in memory mapping data for the given group
::   - has_data (group, result) -- determins if the given group has any mappings
::
:: Metadata actions:
::   - :xcode (status_code) -- displays the message corresponding to the given status code
::   - :about -- displays information about the library. Copyright, license, etc.
::   - :version ([version]) -- returns/displays the version of the library
::
:: Copyright (c) 2010, Rudy Lattae
:: License: Simplified BSD License
:: ============================================================

@echo off
setlocal
shift


:: ============================================================
:: Initialise
:: ============================================================
:lib.init -- global constants, variables and config options
    :: Info -- name, version, copyright, etc.    
    set LIBMAP_INFO_NAME=libmap.cmd
    set LIBMAP_INFO_SHORTNAME=libmap
    set "LIBMAP_INFO_LONGNAME=Mappings Library"
    set LIBMAP_INFO_VERSION=0.9.7b
    set "LIBMAP_INFO_SUMMARY=A Windows batch script helper for managing key:value mappings"
    set "LIBMAP_INFO_COPYRIGHT=Copyright (c) 2010, Rudy Lattae"
    set "LIBMAP_INFO_LICENSE=Simplified BSD License"

    :: Configuration options 
    if "%LIBMAP_OPT_VERBOSE%"=="" set LIBMAP_OPT_VERBOSE="False"

    :: Status (x) codes
    set LIBMAP_X_NO_ACTION_PROVIDED=1
    set LIBMAP_X_INVALID_ACTION=2
    set LIBMAP_X_MISSING_REQUIRED_PARAMETER=3
    set LIBMAP_X_FILE_NOT_FOUND=4


:: ============================================================
:: Front controller
:: ============================================================

:lib.main
    :: make sure we have an action, there is no need to proceed otherwise
    if "%0"=="" (
        if %LIBMAP_OPT_VERBOSE%=="True" call :msg.action_required
        exit /b %LIBMAP_X_NO_ACTION_PROVIDED%
    )
    
    :: capture any global options
    :lib.main._parse_options
    if "%1"=="-v" (set LIBMAP_OPT_VERBOSE="True" & shift /1 & goto lib.main._parse_options)
    
    :: attempt to handle the requested action
    if "%0"=="list" goto do.list
    if "%0"=="get" goto do.get
    if "%0"=="set" goto do.set
    if "%0"=="delete" goto do.delete
    if "%0"=="save" goto do.save
    if "%0"=="load" goto do.load
    if "%0"=="clear" goto do.clear
    if "%0"=="has_data" goto do.has_data
    
    :: process any requests for library meta data
    if "%0"==":xcode" goto meta.xcode
    if "%0"==":about" goto meta.about
    if "%0"==":version" goto meta.version
    
    :: if we get here, we know the action requested is not supported - leave!
    if %LIBMAP_OPT_VERBOSE%=="True" call :msg.invalid_action %0
    
    exit /b %LIBMAP_X_INVALID_ACTION%


:: ============================================================
:: Action handlers
:: ============================================================

:do.list (group, [separator]) -- outputs a list of the mappings in the given group
:: -- group             [in]
:: -- separator         [in, optional, default=" => "] - character/string to use to seperate each key-value pair
    setlocal
    if not "%1"=="" (set "_group=%~1") else (exit /b %LIBMAP_X_MISSING_REQUIRED_PARAMETER%)
    if not "%~2"=="" (set "_separator=%~2") else (set "_separator= =^> ")
    
    (endlocal
        echo.
        for /f "tokens=3* delims=__=" %%a in ('set %_group%') do (
            if %%a NEQ "" echo.%%a%_separator%%%b
        )
    )
    exit /b 0

:do.get (group, key, value) -- retrieves the value associated with a given key
:: -- group     [in]
:: -- key       [in]
:: -- value     [out]
    if not "%1"=="" (set "_group=%~1") else (exit /b %LIBMAP_X_MISSING_REQUIRED_PARAMETER%)
    if not "%2"=="" (set "_key=%~2") else (exit /b %LIBMAP_X_MISSING_REQUIRED_PARAMETER%)
    set "_value="
    
    for /f "tokens=3* delims=__=" %%a in ('set %_group% ') do (
        if %%a==%_key% (set "_value=%%b" & goto map.get._done)
    )
    
    :map.get._done
    (endlocal
        if not "%~3"=="" (set %~3=%_value%) else (echo.%_value%)
    )
    exit /b 0

:do.set (group, key, value) -- maps an key to a given value
:: -- group     [in]
:: -- key       [in]
:: -- value     [in]
    if not "%1"=="" (set "_group=%~1") else (exit /b %LIBMAP_X_MISSING_REQUIRED_PARAMETER%)
    if not "%2"=="" (set "_key=%~2") else (exit /b %LIBMAP_X_MISSING_REQUIRED_PARAMETER%)
    if not "%~3"=="" (set "_value=%~3") else (exit /b %LIBMAP_X_MISSING_REQUIRED_PARAMETER%)

    (endlocal
        set "%_group%_%_key%=%_value%"
    )
    exit /b 0

:do.delete (group, key) -- removes a reference entry from the given map group
:: -- group     [in]
:: -- key       [in]
    if not "%1"=="" (set "_group=%~1") else (exit /b %LIBMAP_X_MISSING_REQUIRED_PARAMETER%)
    if not "%2"=="" (set "_key=%~2") else (exit /b %LIBMAP_X_MISSING_REQUIRED_PARAMETER%)
    
    (endlocal
        set %_group%_%_key%=
    )
    exit /b 0

:do.save (group, data_file_path, [create_file="True"]) -- writes the specified group of mappings to the given file
:: -- group             [in]
:: -- data_file_path    [in]
:: -- create_file       [in, optional, default="True"] -- flag ("True"/"False") if file should be created in case it does not exist yet
    if not "%1"=="" (set "_group=%~1") else (exit /b %LIBMAP_X_MISSING_REQUIRED_PARAMETER%)
    if not "%2"=="" (set "_data_file_path=%~2") else (exit /b %LIBMAP_X_MISSING_REQUIRED_PARAMETER%)
    if not "%3"=="" (set "_create_file=%3") else (set _create_file="True")

    if not exist "%_data_file_path%" (
        if %_create_file%=="True" (call :util.create_file %_data_file_path% & goto :do.save._file_exists)
        if %LIBMAP_OPT_VERBOSE%=="True" (call :msg.file_not_found %_data_file_path%)
        exit /b %LIBMAP_X_FILE_NOT_FOUND%
    )
    
    :do.save._file_exists
    (endlocal
        call :do.list %_group%, ": ">"%_data_file_path%"
        goto :eof
    )
    exit /b 0

:do.load (group, data_file_path) -- loads mappings for group from given data file
:: -- group             [in]
:: -- data_file_path    [in]
    if not "%1"=="" (set "_group=%~1") else (exit /b %LIBMAP_X_MISSING_REQUIRED_PARAMETER%)
    if not "%2"=="" (set "_data_file_path=%~2") else (exit /b %LIBMAP_X_MISSING_REQUIRED_PARAMETER%)

    if not exist "%_data_file_path%" (
        if %LIBMAP_OPT_VERBOSE%=="True" call :msg.file_not_found %_data_file_path%
        exit /b %LIBMAP_X_FILE_NOT_FOUND%
    )

    (endlocal    
        for /f "usebackq tokens=1,* delims=: " %%a in ("%_data_file_path%") do (
            set "%_group%_%%a=%%b"
        )
    )
    exit /b 0

:do.clear (group) -- clears all the in memory mapping data for the given group
:: -- group     [in]
    if not "%1"=="" (set "_group=%~1") else (exit /b %LIBMAP_X_MISSING_REQUIRED_PARAMETER%)
    
    (endlocal
        for /f "tokens=3* delims=__=" %%a in ('set %_group%') do (
            set "%_group%_%%a="
        )
    )
    exit /b 0

:do.has_data (group, [result]) -- determines if the given map group has mappings
:: -- group     [in]
:: -- result    [out, optional, boolean, ("True", "False")]
    if not "%1"=="" (set "_group=%~1") else (exit /b %LIBMAP_X_MISSING_REQUIRED_PARAMETER%)
    set _found="False"
    
    (endlocal
        for /f "tokens=3* delims=__=" %%a in ('set %_group%') do (
            if %%a NEQ "" (set _found="True" & goto :do.has_data._break)
        )
        
        :do.has_data._break
        if not "%~2"=="" (set %~2=%_found%) else (echo.%_found%)
    )
    exit /b 0


:: ============================================================
:: Metadata accessors
:: ============================================================

:meta.xcode (status_code) -- displays the message corresponding to the given status code
    if not "%1"=="" (set "_status_code=%~1") else (goto :eof)
    if "%_status_code%"=="%LIBMAP_X_NO_ACTION_PROVIDED%" (
        echo. No action provided
    )
    if "%_status_code%"=="%LIBMAP_X_INVALID_ACTION%" (
        echo. Invalid action
    )
    if "%_status_code%"=="%LIBMAP_X_MISSING_REQUIRED_PARAMETER%" (
        echo. Missing required parameter
    )
    if "%_status_code%"=="%LIBMAP_X_FILE_NOT_FOUND%" (
        echo. File not found
    )
    goto :eof

:meta.about -- displays information about the library. Copyright, license, etc.
    echo.  %LIBMAP_INFO_LONGNAME% (version %LIBMAP_INFO_VERSION%)
    echo.
    echo.  %LIBMAP_INFO_SUMMARY%
    echo.
    echo.  %LIBMAP_INFO_COPYRIGHT%
    echo.  %LIBMAP_INFO_LICENSE%
    goto :eof

:meta.version ([version]) -- returns/displays the version of the library
:: -- version     [out, optional]
    if "%~1"=="" (echo.  %LIBMAP_INFO_VERSION% & goto :eof)
    (endlocal
        set "%~1=%LIBMAP_INFO_VERSION%"
    )
    goto :eof


:: ============================================================
:: Messages
:: ============================================================

:msg.action_required
    setlocal    
    echo.  Specify an action to perform.
    goto :eof

:msg.invalid_action (action)
    setlocal
    set "_action=%~1"
    echo.  Invalid action "%_action%"
    endlocal
    goto :eof

:msg.file_not_found (file_path)
    setlocal
    set "_file_path=%~1"
    echo.  Cannot locate the file: 
    echo.    %_file_path%
    endlocal
    goto :eof
    
    
:: ============================================================
:: Utilities
:: ============================================================
   
:util.create_file (file_path) -- creates a file if it does not yet exist
:: -- file_path    [in]
    setlocal
    if not "%1"=="" (set "_file_path=%~1") else (goto :eof)
    set "_dir=%~dp1"

    if not exist "%_dir%" mkdir "%_dir%"
    if not exist "%_file_path%" echo.>"%_file_path%"
    endlocal
    goto :eof
